@@ -9,7 +9,7 @@ from uuid import uuid4, UUID
 import pony
 from pony.utils import is_utf8, decorator, throw, localbase, deprecated
 from pony.converting import str2date, str2time, str2datetime, str2timedelta
-from pony.orm.ormtypes import LongStr, LongUnicode, RawSQLType, TrackedValue, Json, QueryType
+from pony.orm.ormtypes import LongStr, LongUnicode, RawSQLType, TrackedValue, TrackedArray, Json, QueryType, Array
 
 class DBException(Exception):
     def __init__(exc, original_exc, *args):
@@ -45,28 +45,43 @@ class     NotSupportedError(DatabaseError): pass
 @decorator
 def wrap_dbapi_exceptions(func, provider, *args, **kwargs):
     dbapi_module = provider.dbapi_module
+    should_retry = False
     try:
-        if provider.dialect != 'SQLite':
-            return func(provider, *args, **kwargs)
-        else:
-            provider.local_exceptions.keep_traceback = True
-            try: return func(provider, *args, **kwargs)
-            finally: provider.local_exceptions.keep_traceback = False
-    except dbapi_module.NotSupportedError as e: raise NotSupportedError(e)
-    except dbapi_module.ProgrammingError as e: raise ProgrammingError(e)
-    except dbapi_module.InternalError as e: raise InternalError(e)
-    except dbapi_module.IntegrityError as e: raise IntegrityError(e)
-    except dbapi_module.OperationalError as e:
-        if provider.dialect == 'SQLite': provider.restore_exception()
-        raise OperationalError(e)
-    except dbapi_module.DataError as e: raise DataError(e)
-    except dbapi_module.DatabaseError as e: raise DatabaseError(e)
-    except dbapi_module.InterfaceError as e:
-        if e.args == (0, '') and getattr(dbapi_module, '__name__', None) == 'MySQLdb':
-            throw(InterfaceError, e, 'MySQL server misconfiguration')
-        raise InterfaceError(e)
-    except dbapi_module.Error as e: raise Error(e)
-    except dbapi_module.Warning as e: raise Warning(e)
+        try:
+            if provider.dialect != 'SQLite':
+                return func(provider, *args, **kwargs)
+            else:
+                provider.local_exceptions.keep_traceback = True
+                try: return func(provider, *args, **kwargs)
+                finally: provider.local_exceptions.keep_traceback = False
+        except dbapi_module.NotSupportedError as e: raise NotSupportedError(e)
+        except dbapi_module.ProgrammingError as e:
+            if provider.dialect == 'PostgreSQL':
+                msg = str(e)
+                if msg.startswith('operator does not exist:') and ' json ' in msg:
+                    msg += ' (Note: use column type `jsonb` instead of `json`)'
+                    raise ProgrammingError(e, msg, *e.args[1:])
+            raise ProgrammingError(e)
+        except dbapi_module.InternalError as e: raise InternalError(e)
+        except dbapi_module.IntegrityError as e: raise IntegrityError(e)
+        except dbapi_module.OperationalError as e:
+            if provider.dialect == 'PostgreSQL' and e.pgcode == '40001':
+                should_retry = True
+            if provider.dialect == 'SQLite':
+                provider.restore_exception()
+            raise OperationalError(e)
+        except dbapi_module.DataError as e: raise DataError(e)
+        except dbapi_module.DatabaseError as e: raise DatabaseError(e)
+        except dbapi_module.InterfaceError as e:
+            if e.args == (0, '') and getattr(dbapi_module, '__name__', None) == 'MySQLdb':
+                throw(InterfaceError, e, 'MySQL server misconfiguration')
+            raise InterfaceError(e)
+        except dbapi_module.Error as e: raise Error(e)
+        except dbapi_module.Warning as e: raise Warning(e)
+    except Exception as e:
+        if should_retry:
+            e.should_retry = True
+        raise
 
 def unexpected_args(attr, args):
     throw(TypeError, 'Unexpected positional argument{} for attribute {}: {}'.format(
@@ -85,13 +100,12 @@ def get_version_tuple(s):
 class DBAPIProvider(object):
     paramstyle = 'qmark'
     quote_char = '"'
-    max_params_count = 200
+    max_params_count = 999
     max_name_len = 128
     table_if_not_exists_syntax = True
     index_if_not_exists_syntax = True
     max_time_precision = default_time_precision = 6
     uint64_support = False
-    select_for_update_nowait_syntax = True
 
     # SQLite and PostgreSQL does not limit varchar max length.
     varchar_default_max_len = None
@@ -101,6 +115,7 @@ class DBAPIProvider(object):
     dbschema_cls = None
     translator_cls = None
     sqlbuilder_cls = None
+    array_converter_cls = None
 
     name_before_table = 'schema_name'
     default_schema_name = None
@@ -270,6 +285,11 @@ class DBAPIProvider(object):
         if isinstance(py_type, type):
             for t, converter_cls in provider.converter_classes:
                 if issubclass(py_type, t): return converter_cls
+            if issubclass(py_type, Array):
+                converter_cls = provider.array_converter_cls
+                if converter_cls is None:
+                    throw(NotImplementedError, 'Array type is not supported for %r' % provider.dialect)
+                return converter_cls
         if isinstance(py_type, RawSQLType):
             return Converter  # for cases like select(raw_sql(...) for x in X)
         throw(TypeError, 'No database converter found for type %s' % py_type)
@@ -511,6 +531,8 @@ class IntConverter(Converter):
         converter.unsigned = unsigned
     def validate(converter, val, obj=None):
         if isinstance(val, int_types): pass
+        elif hasattr(val, '__index__'):
+            val = val.__index__()
         elif isinstance(val, basestring):
             try: val = int(val)
             except ValueError: throw(ValueError,
@@ -647,10 +669,13 @@ class BlobConverter(Converter):
         if isinstance(val, str): return buffer(val)
         throw(TypeError, "Attribute %r: expected type is 'buffer'. Got: %r" % (converter.attr, type(val)))
     def sql2py(converter, val):
-        if not isinstance(val, buffer) or \
-                (PY2 and converter.attr.pk_offset is not None and 'read-write' in repr(val)):  # Issue 355
+        if not isinstance(val, buffer):
             try: val = buffer(val)
             except: pass
+        elif PY2 and converter.attr is not None and converter.attr.is_part_of_unique_index:
+            try: hash(val)
+            except TypeError:
+                val = buffer(val)
         return val
     def sql_type(converter):
         return 'BLOB'
@@ -798,3 +823,50 @@ class JsonConverter(Converter):
         return x == y
     def sql_type(converter):
         return "JSON"
+
+class ArrayConverter(Converter):
+    array_types = {
+        int: ('int', IntConverter),
+        unicode: ('text', StrConverter),
+        float: ('real', RealConverter)
+    }
+
+    def __init__(converter, provider, py_type, attr=None):
+        Converter.__init__(converter, provider, py_type, attr)
+        converter.item_converter = converter.array_types[converter.py_type.item_type][1]
+
+    def validate(converter, val, obj=None):
+        if isinstance(val, TrackedValue) and val.obj_ref() is obj and val.attr is converter.attr:
+            return val
+
+        if isinstance(val, basestring) or not hasattr(val, '__len__'):
+            items = [val]
+        else:
+            items = list(val)
+        item_type = converter.py_type.item_type
+        if item_type == float:
+            item_type = (float, int)
+        for i, v in enumerate(items):
+            if PY2 and isinstance(v, str):
+                v = v.decode('ascii')
+            if not isinstance(v, item_type):
+                if hasattr(v, '__index__'):
+                    items[i] = v.__index__()
+                else:
+                    throw(TypeError, 'Cannot store %s item in array of %s' %
+                          (type(v).__name__, converter.py_type.item_type.__name__))
+
+        if obj is None or converter.attr is None:
+            return items
+        return TrackedArray(obj, converter.attr, items)
+
+    def dbval2val(converter, dbval, obj=None):
+        if obj is None or dbval is None:
+            return dbval
+        return TrackedArray(obj, converter.attr, dbval)
+
+    def val2dbval(converter, val, obj=None):
+        return list(val)
+
+    def sql_type(converter):
+        return '%s[]' % converter.array_types[converter.py_type.item_type][0]
