@@ -3,11 +3,12 @@ from pony.py23compat import PY2, izip, imap, itervalues, basestring, unicode, bu
 
 from operator import attrgetter
 from decimal import Decimal
-from datetime import date, datetime
+from datetime import date, datetime, timedelta
 from binascii import hexlify
 
 from pony import options
 from pony.utils import datetime2timestamp, throw, is_ident
+from pony.converting import timedelta2str
 from pony.orm.ormtypes import RawSQL, Json
 
 class AstError(Exception): pass
@@ -68,19 +69,31 @@ class Value(object):
         self.value = value
     def __unicode__(self):
         value = self.value
-        if value is None: return 'null'
-        if isinstance(value, bool): return value and '1' or '0'
-        if isinstance(value, basestring): return self.quote_str(value)
-        if isinstance(value, datetime): return self.quote_str(datetime2timestamp(value))
-        if isinstance(value, date): return self.quote_str(str(value))
+        if value is None:
+            return 'null'
+        if isinstance(value, bool):
+            return value and '1' or '0'
+        if isinstance(value, basestring):
+            return self.quote_str(value)
+        if isinstance(value, datetime):
+            return 'TIMESTAMP ' + self.quote_str(datetime2timestamp(value))
+        if isinstance(value, date):
+            return 'DATE ' + self.quote_str(str(value))
+        if isinstance(value, timedelta):
+            return "INTERVAL '%s' HOUR TO SECOND" % timedelta2str(value)
         if PY2:
-            if isinstance(value, (int, long, float, Decimal)): return str(value)
-            if isinstance(value, buffer): return "X'%s'" % hexlify(value)
+            if isinstance(value, (int, long, float, Decimal)):
+                return str(value)
+            if isinstance(value, buffer):
+                return "X'%s'" % hexlify(value)
         else:
-            if isinstance(value, (int, float, Decimal)): return str(value)
-            if isinstance(value, bytes): return "X'%s'" % hexlify(value).decode('ascii')
-        assert False, value  # pragma: no cover
-    if not PY2: __str__ = __unicode__
+            if isinstance(value, (int, float, Decimal)):
+                return str(value)
+            if isinstance(value, bytes):
+                return "X'%s'" % hexlify(value).decode('ascii')
+        assert False, repr(value)  # pragma: no cover
+    if not PY2:
+        __str__ = __unicode__
     def __repr__(self):
         return '%s(%r)' % (self.__class__.__name__, self.value)
     def quote_str(self, s):
@@ -97,9 +110,9 @@ def flat(tree):
         x = stack_pop()
         if isinstance(x, basestring): result_append(x)
         else:
-            try: stack_extend(reversed(x))
+            try: stack_extend(x)
             except TypeError: result_append(x)
-    return result
+    return result[::-1]
 
 def flat_conditions(conditions):
     result = []
@@ -254,10 +267,12 @@ class SQLBuilder(object):
             return result
         finally:
             builder.suppress_aliases = prev_suppress_aliases
-    def SELECT_FOR_UPDATE(builder, nowait, *sections):
+    def SELECT_FOR_UPDATE(builder, nowait, skip_locked, *sections):
         assert not builder.indent
         result = builder.SELECT(*sections)
-        return result, 'FOR UPDATE NOWAIT\n' if nowait else 'FOR UPDATE\n'
+        nowait = ' NOWAIT' if nowait else ''
+        skip_locked = ' SKIP LOCKED' if skip_locked else ''
+        return result, 'FOR UPDATE', nowait, skip_locked, '\n'
     def EXISTS(builder, *sections):
         result = builder._subquery(*sections)
         indent = builder.indent_spaces * builder.indent
@@ -457,7 +472,10 @@ class SQLBuilder(object):
         assert distinct in (None, True, False)
         if not distinct:
             if not expr_list: return ['COUNT(*)']
-            return 'COUNT(', join(', ', imap(builder, expr_list)), ')'
+            if builder.dialect == 'PostgreSQL':
+                return 'COUNT(', builder.ROW(*expr_list), ')'
+            else:
+                return 'COUNT(', join(', ', imap(builder, expr_list)), ')'
         if not expr_list: throw(AstError, 'COUNT(DISTINCT) without argument')
         if len(expr_list) == 1:
             return 'COUNT(DISTINCT ', builder(expr_list[0]), ')'
@@ -478,7 +496,10 @@ class SQLBuilder(object):
         assert distinct in (None, True, False)
         result = distinct and 'GROUP_CONCAT(DISTINCT ' or 'GROUP_CONCAT(', builder(expr)
         if sep is not None:
-            result = result, ', ', builder(sep)
+            if builder.provider.dialect == 'MySQL':
+                result = result, ' SEPARATOR ', builder(sep)
+            else:
+                result = result, ', ', builder(sep)
         return result, ')'
     UPPER = make_unary_func('upper')
     LOWER = make_unary_func('lower')
@@ -502,6 +523,93 @@ class SQLBuilder(object):
     def SUBSTR(builder, expr, start, len=None):
         if len is None: return 'substr(', builder(expr), ', ', builder(start), ')'
         return 'substr(', builder(expr), ', ', builder(start), ', ', builder(len), ')'
+    def STRING_SLICE(builder, expr, start, stop):
+        if start is None:
+            start = [ 'VALUE', 0 ]
+
+        if start[0] == 'VALUE':
+            start_value = start[1]
+            if builder.dialect == 'PostgreSQL' and start_value < 0:
+                index_sql = [ 'LENGTH', expr ]
+                if start_value < -1:
+                    index_sql = [ 'SUB', index_sql, [ 'VALUE', -(start_value + 1) ] ]
+            else:
+                if start_value >= 0: start_value += 1
+                index_sql = [ 'VALUE', start_value ]
+        else:
+            inner_sql = start
+            then = [ 'ADD', inner_sql, [ 'VALUE', 1 ] ]
+            else_ = [ 'ADD', [ 'LENGTH', expr ], then ] if builder.dialect == 'PostgreSQL' else inner_sql
+            index_sql = [ 'IF', [ 'GE', inner_sql, [ 'VALUE', 0 ] ], then, else_ ]
+
+        if stop is None:
+            len_sql = None
+        elif stop[0] == 'VALUE':
+            stop_value = stop[1]
+            if start[0] == 'VALUE':
+                start_value = start[1]
+                if start_value >= 0 and stop_value >= 0:
+                    len_sql = [ 'VALUE', stop_value - start_value ]
+                elif start_value < 0 and stop_value < 0:
+                    len_sql = [ 'VALUE', stop_value - start_value ]
+                elif start_value >= 0 and stop_value < 0:
+                    len_sql = [ 'SUB', [ 'LENGTH', expr ], [ 'VALUE', start_value - stop_value ]]
+                    len_sql = [ 'MAX', False, len_sql, [ 'VALUE', 0 ] ]
+                elif start_value < 0 and stop_value >= 0:
+                    len_sql = [ 'SUB', [ 'VALUE', stop_value + 1 ], index_sql ]
+                    len_sql = [ 'MAX', False, len_sql, [ 'VALUE', 0 ] ]
+                else:
+                    assert False  # pragma: nocover1
+            else:
+                start_sql = [ 'COALESCE', start, [ 'VALUE', 0 ] ]
+                if stop_value >= 0:
+                    start_positive = [ 'SUB', stop, start_sql ]
+                    start_negative = [ 'SUB', [ 'VALUE', stop_value + 1 ], index_sql ]
+                else:
+                    start_positive = [ 'SUB', [ 'LENGTH', expr ], [ 'ADD', start_sql, [ 'VALUE', -stop_value ] ] ]
+                    start_negative = [ 'SUB', stop, start_sql]
+                len_sql = [ 'IF', [ 'GE', start_sql, [ 'VALUE', 0 ] ], start_positive, start_negative ]
+                len_sql = [ 'MAX', False, len_sql, [ 'VALUE', 0 ] ]
+        else:
+            stop_sql = [ 'COALESCE', stop, [ 'VALUE', -1 ] ]
+            if start[0] == 'VALUE':
+                start_value = start[1]
+                start_sql = [ 'VALUE', start_value ]
+                if start_value >= 0:
+                    stop_positive = [ 'SUB', stop_sql, start_sql ]
+                    stop_negative = [ 'SUB', [ 'LENGTH', expr ], [ 'SUB', start_sql, stop_sql ] ]
+                else:
+                    stop_positive = [ 'SUB', [ 'ADD', stop_sql, [ 'VALUE', 1 ] ], index_sql ]
+                    stop_negative = [ 'SUB', stop_sql, start_sql]
+                len_sql = [ 'IF', [ 'GE', stop_sql, [ 'VALUE', 0 ] ], stop_positive, stop_negative ]
+                len_sql = [ 'MAX', False, len_sql, [ 'VALUE', 0 ] ]
+            else:
+                start_sql = [ 'COALESCE', start, [ 'VALUE', 0 ] ]
+                both_positive = [ 'SUB', stop_sql, start_sql ]
+                both_negative = both_positive
+                start_positive = [ 'SUB', [ 'LENGTH', expr ], [ 'SUB', start_sql, stop_sql ] ]
+                stop_positive = [ 'SUB', [ 'ADD', stop_sql, [ 'VALUE', 1 ] ], index_sql ]
+                len_sql = [ 'CASE', None, [
+                    (
+                        [ 'AND', [ 'GE', start_sql, [ 'VALUE', 0 ] ], [ 'GE', stop_sql, [ 'VALUE', 0 ] ] ],
+                        both_positive
+                    ),
+                    (
+                        [ 'AND', [ 'LT', start_sql, [ 'VALUE', 0 ] ], [ 'LT', stop_sql, [ 'VALUE', 0 ] ] ],
+                        both_negative
+                    ),
+                    (
+                        [ 'AND', [ 'GE', start_sql, [ 'VALUE', 0 ] ], [ 'LT', stop_sql, [ 'VALUE', 0 ] ] ],
+                        start_positive
+                    ),
+                    (
+                        [ 'AND', [ 'LT', start_sql, [ 'VALUE', 0 ] ], [ 'GE', stop_sql, [ 'VALUE', 0 ] ] ],
+                        stop_positive
+                    ),
+                ]]
+                len_sql = [ 'MAX', False, len_sql, [ 'VALUE', 0 ] ]
+        sql = [ 'SUBSTR', expr, index_sql, len_sql ]
+        return builder(sql)
     def CASE(builder, expr, cases, default=None):
         if expr is None and default is not None and default[0] == 'CASE' and default[1] is None:
             cases2, default2 = default[2:]
@@ -516,6 +624,8 @@ class SQLBuilder(object):
             result.extend((' else ', builder(default)))
         result.append(' end')
         return result
+    def IF(builder, cond, then, else_):
+        return builder.CASE(None, [(cond, then)], else_)
     def TRIM(builder, expr, chars=None):
         if chars is None: return 'trim(', builder(expr), ')'
         return 'trim(', builder(expr), ', ', builder(chars), ')'
@@ -605,3 +715,15 @@ class SQLBuilder(object):
         throw(NotImplementedError)
     def JSON_PARAM(builder, expr):
         return builder(expr)
+    def ARRAY_INDEX(builder, col, index):
+        throw(NotImplementedError)
+    def ARRAY_CONTAINS(builder, key, not_in, col):
+        throw(NotImplementedError)
+    def ARRAY_SUBSET(builder, array1, not_in, array2):
+        throw(NotImplementedError)
+    def ARRAY_LENGTH(builder, array):
+        throw(NotImplementedError)
+    def ARRAY_SLICE(builder, array, start, stop):
+        throw(NotImplementedError)
+    def MAKE_ARRAY(builder, *items):
+        throw(NotImplementedError)
